// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <cutil.h>
#include <multithreading.h>

// Maximum number of CPU threads or GPUs.
#define MAX_CPU_THREAD    4
texture<float, 2> tex3D_LK;// Lookup table mode can be true or false
cudaArray* d_array3D = NULL;

static int s_gpuCount = 0;
void initTextureMemory(int width, int height, int length);
__global__ void cpliGradient_32f(float * xDeriv, float * yDeriv, float * zDeriv, const int width, const int height, const int length);

int flags = 0;
pthread_mutex_t read_mutex    = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t count_mutex    = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t condition_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t count_threshold_cv;

float* h_data;
float* d_data;
float* d_grad;

float* h_result[2];

const int width = 160;
const int height = 192;
const int length = 160;

const int nElems = width * height * length;
const int memSize= nElems * sizeof(float);

//cudaStream_t stream[2];
__device__ __constant__ int d_extInfo[16];
int h_extInfo[16];
int nIters=20;

static CUT_THREADPROC gpuThread(int * device)
{
    CUDA_SAFE_CALL(cudaSetDevice(*device));
    int id = *device;
    initTextureMemory(width, height, length);

    fprintf(stderr, "Device %d \n", id);
    cudaMalloc((void**)&d_data, memSize);
    cudaMalloc((void**)&d_grad, 3 * memSize);
    
    cudaMemcpy(d_data, h_data, memSize, cudaMemcpyHostToDevice);
    cudaMemcpyToArray(d_array3D, 0,0 , d_data, memSize, cudaMemcpyDeviceToDevice);

    dim3 threads(16, 16);
    dim3 grids(width / 16, height/16);
    for (unsigned int i=0; i <nIters; ++i) {
        fprintf(stderr, "Device %d Iter %d\n", id, i);
        cpliGradient_32f<<<threads, grids>>>(d_grad, d_grad + nElems, d_grad + 2 * nElems, width , height, length);
    }
    cudaMemcpy(h_result[id], d_grad, memSize * 3, cudaMemcpyDeviceToHost);
                
    CUT_THREADEND;
}

float imgSum(float* data, int len){
    float sum=0;
    for (int i=0; i< len; ++i)
        sum+=data[i];
    return sum;
}

float imgSum2(float* data, int len){
    float sum=0;
    for (int i=0; i< len; ++i)
        sum+=data[i] * data[i];
    return sum;
}

void createRandomData(float* data, float len){
    for (unsigned int i=0; i<len; ++i)
        data[i] = (float) rand() / RAND_MAX;
}
////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int main( int argc, char** argv)
{
    CUT_DEVICE_INIT();  
    CUDA_SAFE_CALL(cudaGetDeviceCount(&s_gpuCount));
    cutGetCmdLineArgumenti(argc, (const char**) argv, "nIters", &nIters);
 
    h_data = (float*) malloc(memSize);
    createRandomData(h_data, nElems);
    
    h_result[0] = (float*) malloc(3 * memSize);
    h_result[1] = (float*) malloc(3 * memSize);

    unsigned int timer = 0;
    CUT_SAFE_CALL(cutCreateTimer(&timer));
    int threadIds[MAX_CPU_THREAD];
    printf("%d GPUs found\n", s_gpuCount);
    CUT_SAFE_CALL(cutStartTimer(timer));
    CUTThread * threads = (CUTThread *)malloc(sizeof(CUTThread) * s_gpuCount);
    
    for(int i = 0; i < s_gpuCount; i++)
    {
        threadIds[i] = i;
        threads[i] = cutStartThread((CUT_THREADROUTINE)gpuThread, (void *)&threadIds[i]);
    }

    // Wait for all the threads to finish.
    cutWaitForThreads(threads, s_gpuCount);
    // Test the result
    int test = 1;
    float sum= imgSum2(h_result[0], nElems);
    fprintf(stderr," Sum %f \n", sum);
    float sum2= imgSum2(h_result[1], nElems);
    fprintf(stderr," Sum %f \n", sum2);
    for (int i=0; i< nElems * 3; ++i)
        if (h_result[0][i] != h_result[1][i]){
            test = 0;
            break;
        }

    if (!test) 
        fprintf(stderr, "TEST FAILED");
    else
        fprintf(stderr, "TEST PASSED");

    free(threads);
    CUT_SAFE_CALL(cutStopTimer(timer));
    printf("Processing time: %f (ms)\n", cutGetTimerValue(timer));
    CUT_SAFE_CALL(cutDeleteTimer(timer));
    CUT_EXIT(argc, argv);

    free(h_data);
}

__global__ void cpliGradient_32f(float * xDeriv, float * yDeriv, float * zDeriv,
                                 const int width, const int height, const int length)
{

    int i = threadIdx.x + blockIdx.x * blockDim.x;
    int j = threadIdx.y + blockIdx.y * blockDim.y;
    int index = i + j * width;
    const int wh = width * height;

    float xyz, x0yz, x1yz, xy0z, xy1z, xyz0, xyz1;
    float x, y;

    x = (float)i + 0.5f;
    y = (float)j + 0.5f;

    xyz  = 0;
    xyz1 = tex2D(tex3D_LK, x, y);
    for (int k=0 ; k < length; ++k, index += wh){
        x1yz = tex2D(tex3D_LK, x + 1.f, y);
        x0yz = tex2D(tex3D_LK, x - 1.f, y);
        xy1z = tex2D(tex3D_LK, x, y + 1.f);
        xy0z = tex2D(tex3D_LK, x, y - 1.f);

        xyz0 = xyz;
        xyz  = xyz1;

        y += height;
        xyz1 = tex2D(tex3D_LK, x,y);

        xDeriv[index] = 0.5f * (x1yz - x0yz);
        yDeriv[index] = 0.5f * (xy1z - xy0z);
        zDeriv[index] = 0.5f * (xyz1 - xyz0);
    }

}


void initTextureMemory(int width, int height, int length){

    //3 Prepare texture memory for memory trilinear interpolation 
    tex3D_LK.addressMode[0] = cudaAddressModeClamp;
    tex3D_LK.addressMode[1] = cudaAddressModeClamp;
    tex3D_LK.filterMode     = cudaFilterModePoint;
    tex3D_LK.normalized     = false;

    cudaChannelFormatDesc desc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);
    CUDA_SAFE_CALL(cudaMallocArray(&d_array3D, &desc, width, height * length));
    cudaBindTextureToArray(tex3D_LK, d_array3D);    
}


